package net.codinux.banking.client.fints4k

import net.codinux.banking.client.BankingClient
import net.codinux.banking.client.BankingClientCallback
import net.codinux.banking.client.model.BankAccount
import net.codinux.banking.client.model.BankAccountFeatures
import net.codinux.banking.client.model.BankAccess
import net.codinux.banking.client.model.options.GetAccountDataOptions
import net.codinux.banking.client.model.request.GetAccountDataRequest
import net.codinux.banking.client.model.request.TransferMoneyRequestForUser
import net.codinux.banking.client.model.response.*
import net.codinux.banking.fints.FinTsClient
import net.codinux.banking.fints.config.FinTsClientConfiguration
import net.codinux.banking.fints.model.BankData

open class FinTs4kBankingClient(
    config: FinTsClientConfiguration = FinTsClientConfiguration(),
    callback: BankingClientCallback
) : BankingClient {

    constructor(callback: BankingClientCallback) : this(FinTsClientConfiguration(), callback)

    constructor(options: FinTsClientOptions, callback: BankingClientCallback)
            : this(FinTsClientConfiguration(net.codinux.banking.fints.config.FinTsClientOptions(options.collectMessageLog, false, options.removeSensitiveDataFromMessageLog, options.closeDialogs, options.version, options.productName)), callback)


    protected open val mapper = FinTs4kMapper()

    protected open val client = FinTsClient(config, BridgeFintTsToBankingClientCallback(callback, mapper))


    override suspend fun getAccountDataAsync(request: GetAccountDataRequest): Response<GetAccountDataResponse> {
        val response = client.getAccountDataAsync(mapper.mapToGetAccountDataParameter(request, request.options ?: GetAccountDataOptions()))

        return mapper.map(response, request.bankInfo)
    }

    override suspend fun updateAccountTransactionsAsync(bank: BankAccess, accounts: List<BankAccount>?): Response<List<GetTransactionsResponse>> {
        val accountsToRequest = (accounts ?: bank.accounts).filter { it.supportsAnyFeature(BankAccountFeatures.RetrieveBalance, BankAccountFeatures.RetrieveBalance) }

        if (accountsToRequest.isNotEmpty()) {
            var finTsModel: BankData? = null

            val responses = accountsToRequest.map { account ->
                val parameter = mapper.mapToUpdateAccountTransactionsParameter(bank, account, finTsModel)

                val response = client.getAccountDataAsync(parameter)

                if (response.finTsModel != null) {
                    finTsModel = response.finTsModel // so that basic account data doesn't have to be retrieved another time if user has multiple accounts
                }

                Triple(account, parameter, response)
            }

            return mapper.map(responses)
        }

        return Response.error(ErrorType.NoneOfTheAccountsSupportsRetrievingData, "Keiner der Konten unterstützt das Abholen der Umsätze oder des Kontostands") // TODO: translate
    }


    override suspend fun transferMoneyAsync(request: TransferMoneyRequestForUser): Response<TransferMoneyResponse> {
        val response = client.transferMoneyAsync(mapper.mapToTransferMoneyParameter(request))

        return mapper.mapTransferMoneyResponse(response)
    }

}