package net.codinux.banking.client.fints4k

import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import kotlinx.datetime.TimeZone
import kotlinx.datetime.toLocalDateTime
import net.codinux.banking.client.model.*
import net.codinux.banking.client.model.AccountTransaction
import net.codinux.banking.client.model.Amount
import net.codinux.banking.client.model.extensions.EuropeBerlin
import net.codinux.banking.client.model.tan.*
import net.codinux.banking.client.model.options.GetAccountDataOptions
import net.codinux.banking.client.model.request.GetAccountDataRequest
import net.codinux.banking.client.model.request.TransferMoneyRequestForUser
import net.codinux.banking.client.model.response.*
import net.codinux.banking.client.model.tan.ActionRequiringTan
import net.codinux.banking.client.model.tan.TanChallenge
import net.codinux.banking.client.model.tan.TanImage
import net.codinux.banking.client.model.tan.TanMethod
import net.codinux.banking.client.model.tan.TanMethodType
import net.dankito.banking.client.model.BankAccountIdentifierImpl
import net.dankito.banking.client.model.parameter.GetAccountDataParameter
import net.dankito.banking.client.model.parameter.RetrieveTransactions
import net.dankito.banking.client.model.response.ErrorCode
import net.codinux.banking.fints.mapper.FinTsModelMapper
import net.codinux.banking.fints.messages.datenelemente.implementierte.signatur.Sicherheitsfunktion
import net.codinux.banking.fints.model.*
import net.codinux.banking.fints.transactions.swift.model.Holding
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.TanMedium
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.MobilePhoneTanMedium
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.TanGeneratorTanMedium
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.TanMediumStatus
import net.dankito.banking.banklistcreator.prettifier.BankingGroupMapper
import net.dankito.banking.client.model.parameter.TransferMoneyParameter
import kotlin.io.encoding.Base64
import kotlin.io.encoding.ExperimentalEncodingApi

open class FinTs4kMapper {

    protected val fintsModelMapper = FinTsModelMapper()

    protected val bankingGroupMapper = BankingGroupMapper()


    open fun mapToGetAccountDataParameter(request: GetAccountDataRequest, options: GetAccountDataOptions) = GetAccountDataParameter(
        request.bankCode, request.loginName, request.password,
        options.accounts.map { mapBankAccountIdentifier(it) },
        options.retrieveBalance,
        RetrieveTransactions.valueOf(options.retrieveTransactions.name), options.retrieveTransactionsFrom, options.retrieveTransactionsTo,
        preferredTanMethods = options.preferredTanMethods?.map { mapTanMethodType(it) },
        tanMethodsNotSupportedByApplication = options.tanMethodsNotSupportedByApplication.map { mapTanMethodType(it) },
        abortIfTanIsRequired = options.abortIfTanIsRequired,
        defaultBankValues = request.bankInfo?.let { mapToBankData(request, it) }
    )

    protected open fun mapToBankData(credentials: AccountCredentials, bank: BankInfo): BankData = BankData(
        credentials.bankCode, credentials.loginName, credentials.password,
        bank.serverAddress, bank.bic, bank.name
    )

    open fun mapToUpdateAccountTransactionsParameter(bank: BankAccess, account: BankAccount, finTsModel: BankData?): GetAccountDataParameter {
        val defaults = GetAccountDataOptions()

        val accountIdentifier = BankAccountIdentifierImpl(account.identifier, account.subAccountNumber, account.iban)
        val from = account.lastAccountUpdateTime?.toLocalDateTime(TimeZone.EuropeBerlin)?.date // TODO: in case lastTransactionsUpdateTime is not set, this would retrieve all transactions (and require a TAN im most cases)
        val retrieveTransactions = if (from != null) RetrieveTransactions.AccordingToRetrieveFromAndTo else RetrieveTransactions.valueOf(defaults.retrieveTransactions.name)
//        val preferredTanMethods = listOf(mapTanMethodType(bank.selectedTanMethod.type)) // TODO: currently we aren't saving TanMethods in database, re-enable as soon as TanMethods get saved
        val preferredTanMethods = defaults.preferredTanMethods?.map { mapTanMethodType(it) }

        return GetAccountDataParameter(bank.domesticBankCode, bank.loginName, bank.password!!, listOf(accountIdentifier), true,
            retrieveTransactions, from,
            preferredTanMethods = preferredTanMethods,
            preferredTanMedium = bank.selectedTanMediumIdentifier,
            finTsModel = finTsModel
        )
    }

    open fun mapBankAccountIdentifier(account: BankAccountIdentifier): BankAccountIdentifierImpl =
        BankAccountIdentifierImpl(account.identifier, account.subAccountNumber, account.iban)

    open fun mapTanMethod(method: TanMethod) = net.codinux.banking.fints.model.TanMethod( // TODO: get instance from FinTsData, don't create manually
        method.displayName, Sicherheitsfunktion.entries.first { it.code == method.identifier }, mapTanMethodType(method.type), null, method.maxTanInputLength, mapAllowedTanFormat(method.allowedTanFormat)
    )

    protected open fun mapTanMethodType(type: TanMethodType): net.codinux.banking.fints.model.TanMethodType =
        net.codinux.banking.fints.model.TanMethodType.valueOf(type.name)

    protected open fun mapAllowedTanFormat(allowedTanFormat: AllowedTanFormat?): net.codinux.banking.fints.messages.datenelemente.implementierte.tan.AllowedTanFormat =
        allowedTanFormat?.let { net.codinux.banking.fints.messages.datenelemente.implementierte.tan.AllowedTanFormat.valueOf(it.name) } ?: net.codinux.banking.fints.messages.datenelemente.implementierte.tan.AllowedTanFormat.Alphanumeric


    open fun map(response: net.dankito.banking.client.model.response.GetAccountDataResponse, bank: BankInfo? = null): Response<GetAccountDataResponse> =
        if (response.successful && response.customerAccount != null) {
            Response.success(GetAccountDataResponse(mapBank(response.customerAccount!!, bank)))
        } else {
            mapError(response)
        }

    open fun map(responses: List<Triple<BankAccount, GetAccountDataParameter, net.dankito.banking.client.model.response.GetAccountDataResponse>>): Response<List<GetTransactionsResponse>> {
        val type = if (responses.all { it.third.successful }) ResponseType.Success else ResponseType.Error

        // TODO: update BankAccess and BankAccount objects according to retrieved data
        val mappedResponses = responses.map { (account, param, getAccountDataResponse) ->
            val bank = getAccountDataResponse.customerAccount
            val finTsBankAccount = bank?.accounts?.firstOrNull { it.identifier == account.identifier && it.subAccountNumber == account.subAccountNumber }

            if (getAccountDataResponse.successful && bank != null && finTsBankAccount != null) {
                if (finTsBankAccount.lastAccountUpdateTime != null) {
                    account.lastAccountUpdateTime = finTsBankAccount.lastAccountUpdateTime
                }
                if (account.retrievedTransactionsFrom == null || (finTsBankAccount.retrievedTransactionsFrom != null
                            && account.retrievedTransactionsFrom!! < finTsBankAccount.retrievedTransactionsFrom!!)) {
                    account.retrievedTransactionsFrom = finTsBankAccount.retrievedTransactionsFrom
                }

                Response.success(GetTransactionsResponse(account, mapMoney(finTsBankAccount.balance), mapBookedTransactions(finTsBankAccount), emptyList(),
                    mapHoldings(finTsBankAccount.statementOfHoldings, finTsBankAccount.currency, finTsBankAccount.lastAccountUpdateTime),
                    finTsBankAccount.lastAccountUpdateTime ?: Clock.System.now(), param.retrieveTransactionsFrom, param.retrieveTransactionsTo))
            } else {
                mapError(getAccountDataResponse)
            }
        }

        val data = mappedResponses.filter { it.type == ResponseType.Success }.mapNotNull { it.data }

        return (object : Response<List<GetTransactionsResponse>>(type, data, mappedResponses.firstNotNullOfOrNull { it.error }) { })
    }


    open fun mapToBankViewInfo(bank: BankData): BankViewInfo = BankViewInfo(
        bank.bankCode, bank.customerId, bank.bankName, getBankingGroup(bank.bankName, bank.bic)
    )

    open fun mapToBankAccountViewInfo(account: AccountData): BankAccountViewInfo = BankAccountViewInfo(
        account.accountIdentifier, account.subAccountAttribute,
        mapAccountType(fintsModelMapper.map(account.accountType)),
        account.iban, account.productName
    )


    protected open fun mapBank(bank: net.dankito.banking.client.model.CustomerAccount, info: BankInfo? = null) = BankAccess(
        bank.bankCode, bank.loginName, bank.password,
        info?.name ?: bank.bankName, bank.bic, bank.customerName, bank.userId,
        bank.accounts.map { mapAccount(it) }.sortedBy { it.type }
            .onEachIndexed { index, bankAccount -> bankAccount.displayIndex = index },

        bank.selectedTanMethod?.securityFunction?.code, bank.tanMethods.map { mapTanMethod(it) }.toMutableList(),
        bank.selectedTanMedium?.mediumName, bank.tanMedia.map { mapTanMedium(it) }.toMutableList(),

        info?.bankingGroup ?: getBankingGroup(bank.bankName, bank.bic),
        bank.finTsServerAddress,
        "de"
    )

    protected open fun getBankingGroup(bankName: String, bic: String): BankingGroup? =
        bankingGroupMapper.getBankingGroup(bankName, bic)


    protected open fun mapAccount(account: net.dankito.banking.client.model.BankAccount) = BankAccount(
        account.identifier, account.subAccountNumber, account.iban, account.productName, account.accountHolderName,
        mapAccountType(account.type), account.currency, account.accountLimit,
        account.isAccountTypeSupportedByApplication, mapFeatures(account),
        mapMoney(account.balance),
        account.serverTransactionsRetentionDays,
        account.lastAccountUpdateTime, account.retrievedTransactionsFrom,
        bookedTransactions = mapBookedTransactions(account).toMutableList(),
        holdings = mapHoldings(account.statementOfHoldings, account.currency, account.lastAccountUpdateTime).toMutableList()
    )

    protected open fun mapAccountType(type: net.dankito.banking.client.model.BankAccountType): BankAccountType =
        BankAccountType.valueOf(type.name)

    protected open fun mapFeatures(account: net.dankito.banking.client.model.BankAccount): Set<BankAccountFeatures> = buildSet {
        if (account.supportsRetrievingBalance) {
            add(BankAccountFeatures.RetrieveBalance)
        }
        if (account.supportsRetrievingTransactions) {
            add(BankAccountFeatures.RetrieveTransactions)
        }
        if (account.supportsTransferringMoney) {
            add(BankAccountFeatures.TransferMoney)
        }
        if (account.supportsInstantPayment) {
            add(BankAccountFeatures.InstantTransfer)
        }
    }


    protected open fun mapBookedTransactions(account: net.dankito.banking.client.model.BankAccount): List<AccountTransaction> =
        account.bookedTransactions.map { mapTransaction(it) }

    protected open fun mapTransaction(transaction: net.dankito.banking.client.model.AccountTransaction): AccountTransaction = AccountTransaction(
        mapMoney(transaction.amount), transaction.amount.currency.code, transaction.reference,
        transaction.bookingDate, transaction.valueDate,
        transaction.otherPartyName, transaction.otherPartyBankId, transaction.otherPartyAccountId,

        transaction.postingText,
        mapNullableMoney(transaction.openingBalance), mapNullableMoney(transaction.closingBalance),

        transaction.statementNumber, transaction.sheetNumber,

        transaction.customerReference, transaction.bankReference,
        transaction.furtherInformation,

        transaction.endToEndReference, transaction.mandateReference,
        transaction.creditorIdentifier, transaction.originatorsIdentificationCode,
        transaction.compensationAmount, transaction.originalAmount,
        transaction.deviantOriginator, transaction.deviantRecipient,
        transaction.referenceWithNoSpecialType,

        transaction.journalNumber, transaction.textKeyAddition,

        transaction.orderReferenceNumber, transaction.referenceNumber,
        transaction.isReversal
    )

    protected open fun mapHoldings(statements: List<net.codinux.banking.fints.transactions.swift.model.StatementOfHoldings>, accountCurrency: String, lastAccountUpdateTime: Instant? = null) =
        statements.flatMap { mapHoldings(it, accountCurrency, lastAccountUpdateTime) }

    protected open fun mapHoldings(statement: net.codinux.banking.fints.transactions.swift.model.StatementOfHoldings, accountCurrency: String, lastAccountUpdateTime: Instant? = null): List<net.codinux.banking.client.model.securitiesaccount.Holding> {

        val totalBalance = mapNullableAmount(statement.totalBalance)
        val currency = statement.currency ?: accountCurrency
        val statementDate: Instant? = /* statement.statementDate ?: statement.preparationDate ?: */ lastAccountUpdateTime // TODO

        return statement.holdings.map { mapHolding(it, currency, statementDate, if (statement.holdings.size == 1) totalBalance else null) }
    }

    protected open fun mapHolding(holding: Holding, accountCurrency: String, statementDate: Instant?, totalBalance: Amount? = null) = net.codinux.banking.client.model.securitiesaccount.Holding(
        holding.name, holding.isin, holding.wkn,

        holding.quantity, holding.currency ?: accountCurrency,

        getTotalBalance(holding), mapNullableAmount(holding.marketValue),
        calculatePerformance(holding),
        getTotalCostPrice(holding), mapNullableAmount(holding.averageCostPrice),

        holding.pricingTime ?: statementDate, holding.buyingDate
    )

    // visible for testing
    internal fun getTotalBalance(holding: Holding): Amount? {
        return if (holding.totalBalance != null) {
            mapNullableAmount(holding.totalBalance)
        } else if (holding.quantity != null && holding.marketValue != null) {
            Amount((holding.quantity!! * holding.marketValue.toString().toDouble()).toString())
        } else {
            null
        }
    }

    // visible for testing
    internal fun getTotalCostPrice(holding: Holding): Amount? {
        return if (holding.totalCostPrice != null) {
            mapNullableAmount(holding.totalCostPrice)
        } else if (holding.quantity != null && holding.averageCostPrice != null) {
            Amount((holding.quantity!! * holding.averageCostPrice.toString().toDouble()).toString())
        } else {
            null
        }
    }

    // visible for testing
    internal fun calculatePerformance(holding: Holding): Float? {
        val totalBalance = getTotalBalance(holding)
        val totalCostPrice = getTotalCostPrice(holding)

        if (totalBalance != null && totalCostPrice != null) {
            return ((totalBalance - totalCostPrice) / totalCostPrice).toFloat() * 100
        }

        val marketValue = mapNullableAmount(holding.marketValue)
        val costPrice = mapNullableAmount(holding.averageCostPrice)
        if (marketValue != null && costPrice != null) {
            return ((marketValue - costPrice) / costPrice).toFloat() * 100
        }

        return null
    }

    protected open fun mapNullableMoney(amount: Money?) = amount?.let { mapMoney(it) }

    protected open fun mapMoney(amount: Money) = Amount(amount.amount.string.replace(',', '.'))

    protected open fun mapNullableAmount(amount: net.codinux.banking.fints.model.Amount?) = amount?.let { mapAmount(it) }

    protected open fun mapAmount(amount: net.codinux.banking.fints.model.Amount) = Amount(amount.string.replace(',', '.'))


    open fun mapTanChallenge(challenge: net.codinux.banking.fints.model.TanChallenge): TanChallenge {
        val type = mapTanChallengeType(challenge)
        val action = mapActionRequiringTan(challenge.forAction)

        val tanMethods = challenge.bank.tanMethodsAvailableForUser.map { mapTanMethod(it) }
        val selectedTanMethodId = challenge.tanMethod.securityFunction.code

        val tanMedia = challenge.bank.tanMedia.map { mapTanMedium(it) }
        // TanMedium has not natural id in FinTS model so we have to create our own one
        val selectedTanMediumName = challenge.bank.selectedTanMedium?.let { selected -> tanMedia.firstOrNull { it == selected } }?.identifier

        val bank = mapToBankViewInfo(challenge.bank)
        val account = challenge.account?.let { mapToBankAccountViewInfo(it) }

        val tanImage = if (challenge is ImageTanChallenge) mapTanImage(challenge.image) else null
        val flickerCode = if (challenge is FlickerCodeTanChallenge) mapFlickerCode(challenge.flickerCode) else null

        return object : TanChallenge(type, action, challenge.messageToShowToUser, selectedTanMethodId, tanMethods, selectedTanMediumName, tanMedia, tanImage, flickerCode, bank, account, challenge.tanExpirationTime, challenge.challengeCreationTimestamp) {
            override fun addTanExpiredCallback(callback: () -> Unit) {
                challenge.addTanExpiredCallback(callback)
            }

            override fun addUserApprovedDecoupledTanCallback(callback: () -> Unit) {
                challenge.addUserApprovedDecoupledTanCallback(callback)
            }
        }
    }

    protected open fun mapTanChallengeType(challenge: net.codinux.banking.fints.model.TanChallenge): TanChallengeType = when {
        challenge is ImageTanChallenge -> TanChallengeType.Image
        challenge is FlickerCodeTanChallenge -> TanChallengeType.Flickercode
        else -> TanChallengeType.EnterTan
    }

    protected open fun mapActionRequiringTan(action: net.codinux.banking.fints.model.ActionRequiringTan): ActionRequiringTan =
        ActionRequiringTan.valueOf(action.name)

    open fun mapTanMethod(method: net.codinux.banking.fints.model.TanMethod): TanMethod = TanMethod(
        method.displayName, mapTanMethodType(method.type), method.securityFunction.code, method.maxTanInputLength, mapAllowedTanFormat(method.allowedTanFormat)
    )

    protected open fun mapTanMethodType(type: net.codinux.banking.fints.model.TanMethodType): TanMethodType =
        TanMethodType.valueOf(type.name)

    protected open fun mapAllowedTanFormat(allowedTanFormat: net.codinux.banking.fints.messages.datenelemente.implementierte.tan.AllowedTanFormat?): AllowedTanFormat =
        allowedTanFormat?.let { AllowedTanFormat.valueOf(it.name) } ?: AllowedTanFormat.Alphanumeric

    protected open fun mapTanImage(image: net.codinux.banking.fints.tan.TanImage): TanImage =
        TanImage(image.mimeType, mapToBase64(image.imageBytes), mapException(image.decodingError))

    @OptIn(ExperimentalEncodingApi::class)
    protected open fun mapToBase64(bytes: ByteArray): String {
        return Base64.Default.encode(bytes)
    }

    protected open fun mapTanMedium(tanMedium: TanMedium) = net.codinux.banking.client.model.tan.TanMedium(
        mapTanMediumType(tanMedium), tanMedium.mediumName, mapTanMediumStatus(tanMedium.status),
        (tanMedium as? TanGeneratorTanMedium)?.let { mapTanGeneratorTanMedium(it) },
        (tanMedium as? MobilePhoneTanMedium)?.let { mapMobilePhoneTanMedium(it) }
    )

    protected open fun mapTanMediumStatus(status: TanMediumStatus): net.codinux.banking.client.model.tan.TanMediumStatus = when (status) {
        TanMediumStatus.Aktiv -> net.codinux.banking.client.model.tan.TanMediumStatus.Used
        TanMediumStatus.Verfuegbar -> net.codinux.banking.client.model.tan.TanMediumStatus.Available
        TanMediumStatus.AktivFolgekarte -> net.codinux.banking.client.model.tan.TanMediumStatus.ActiveFollowUpCard
        TanMediumStatus.VerfuegbarFolgekarte -> net.codinux.banking.client.model.tan.TanMediumStatus.AvailableFollowUpCard
    }

    protected open fun mapMobilePhoneTanMedium(tanMedium: MobilePhoneTanMedium) = net.codinux.banking.client.model.tan.MobilePhoneTanMedium(
        tanMedium.phoneNumber, tanMedium.concealedPhoneNumber
    )

    protected open fun mapTanGeneratorTanMedium(tanMedium: TanGeneratorTanMedium) = net.codinux.banking.client.model.tan.TanGeneratorTanMedium(
        tanMedium.cardNumber, tanMedium.cardSequenceNumber, tanMedium.cardType,
        tanMedium.validFrom, tanMedium.validTo
    )

    protected open fun mapTanMediumType(tanMedium: TanMedium): TanMediumType = when {
        tanMedium is MobilePhoneTanMedium -> TanMediumType.MobilePhone
        tanMedium is TanGeneratorTanMedium -> TanMediumType.TanGenerator
        else -> TanMediumType.Generic
    }

    protected open fun mapFlickerCode(flickerCode: net.codinux.banking.fints.tan.FlickerCode): FlickerCode =
        FlickerCode(flickerCode.challengeHHD_UC, flickerCode.parsedDataSet, mapException(flickerCode.decodingError))


    /*      Transfer Money      */

    open fun mapToTransferMoneyParameter(request: TransferMoneyRequestForUser): TransferMoneyParameter = TransferMoneyParameter(
        request.bankCode, request.loginName, request.password, request.senderAccount?.let { mapBankAccountIdentifier(it) },
        request.recipientName, request.recipientAccountIdentifier, request.recipientBankIdentifier,
        mapToMoney(request.amount, request.currency), request.paymentReference, request.instantTransfer,
        request.preferredTanMethods?.map { mapTanMethodType(it) },
        request.tanMethodsNotSupportedByApplication.map { mapTanMethodType(it) }
    )

    open fun mapTransferMoneyResponse(response: net.dankito.banking.client.model.response.TransferMoneyResponse): Response<TransferMoneyResponse>  =
        if (response.successful) {
            Response.success(TransferMoneyResponse())
        } else {
            mapError(response)
        }

    open fun mapToMoney(amount: Amount, currency: String): Money = Money(amount.toString(), currency)


    protected open fun <T> mapError(response: net.dankito.banking.client.model.response.FinTsClientResponse): Response<T> {
        return if (response.error != null) {
            Response.error(ErrorType.valueOf(response.error!!.name), if (response.error == ErrorCode.BankReturnedError) null else response.errorMessage,
                if (response.error == ErrorCode.BankReturnedError && response.errorMessage !== null) listOf(response.errorMessage!!) else emptyList())
        } else {
            Response.error(ErrorType.UnknownError, response.errorMessage)
        }
    }

    protected open fun mapException(exception: Exception?): String? =
        exception?.stackTraceToString()

}